* ta 12, 20, 40, 72; si1; sl1; sr80

*======================================================================*
* The BLACK_HOLE - my very first device driver !                       *
*                                                                      *
* Works like a dustbin for listing, saving, printing to etc, in a way  *
* like the one in Quanta January/ February 1989. Similar to the 'nul'  *
* device in Lightning except this one calls te correct allocation      *
* vectors ! (according to QDOS documentation anyway...                 *
*----------------------------------------------------------------------*
* As with all normal black holes, anything goes in, but nothing gets   *
* out. You can PRINT, SAVE, LIST, SPL, DIR, BPUT etc to one but if you *
* try to INPUT, INKEY$, LOAD etc you will get END_OF_FILE.             *
*                                                                      *
* Anybody trying to CLS, or PAPER etc a BLACK_HOLE channel gets BAD    *
* PARAMETER and deservedly so.                                         *
*----------------------------------------------------------------------*
* The following list gives details of what is allowed in QDOS speak    *
*                                                                      *
* TRAP #1 - not applicable.                                            *
* TRAP #2 - IO_OPEN & IO_CLOSE                        both work ok.    *
* TRAP #3 - IO_SBYTE, FS_CHECK, FS_FLUSH, FS_POSAB, FS_POSRE           *
*           IO_SSTRG, FS_SAVE                         all work ok.     *
*           IO_PEND, IO_FBYTE, IO_FLINE, IO_FSTRG, IO_EDLIN            *
*           FS_MDINF, FS_HEADR, FS_LOAD               all give ERR_EF. *
*           All other D0 codes give ERR_BP.                            *
*                                                                      *
* Note that IO_SSTRG & FS_SAVE return correct parameters as if they    *
* had done something, ie, A1.L = A1.L + D2.W, D1.W = D2.W              *
*======================================================================*
* Copyright Norman Dunbar                                              *
*           February 1991                                              *
*======================================================================*

           nolist
           get     "win2_trap1_hdr"
           get     "win2_trap2_hdr"
           get     "win2_trap3_hdr"
           get     "win2_vectors_hdr"
           get     "win2_errors_hdr"
           list

*----------------------------------------------------------------------*
* Initialise the driver by setting up its linkage block and filling in *
* the addresses of the OPEN, CLOSE and INPUT/OUTPUT routines. Then it  *
* is linked into the (simple) device driver linked list.               *
*----------------------------------------------------------------------*

start      lea.l   drvr_block,a0       linkage block address
           lea.l   dev_io,a1           I/O routine address
           move.l  a1,4(a0)            Into its slot
           lea.l   dev_open,a1         OPEN routine address
           move.l  a1,8(a0)            Into its slot
           lea.l   dev_close,a1        CLOSE routine address
           move.l  a1,$0c(a0)          Into its slot
           moveq   #mt_liod,d0         Now prepare to ...
           trap    #1                  ... link driver into QDOS and
           rts                         exit with any error codes.

*----------------------------------------------------------------------*
* Four long words for the linkage block. The first is the link to the  *
* next driver in the list, then the INPUT/OUTPUT procedure address,    *
* the OPEN and finally the CLOSE procedure addresses.                  *
*----------------------------------------------------------------------*

drvr_block ds.l    4

*----------------------------------------------------------------------*
* The close routine is simple. A0.L is the address of the channel      *
* definition block which is simply reclaimed using the proper vector   *
* as per the QDOS documentation, and not as per DIGITAL PRECISION !!   *
*----------------------------------------------------------------------*

dev_close  move.w  mm_rechp,a2         Reclaim heap space
           jsr     (a2)                And do it
           moveq   #0,d0               CLOSE never fails
           rts                         And exit

*----------------------------------------------------------------------*
* The OPEN routine decodes the device name and if correct, claims some *
* memory from the heap for a channel definition block. 24 bytes is the *
* minimum size required by QDOS and is filled in by QDOS for us.       *
*----------------------------------------------------------------------*

dev_open   move.w  io_name,a2          Utility to check device names
           jsr     (a2)                And do it
           rts                         Oops, not found error return
           rts                         Oops, bad parameter error return
           bra.s   dev_ok              Yippee !
           dc.w    10                  Size of device name
           dc.b    'BLACK_HOLE'        Device name
           dc.w    0                   Has no parameters

dev_ok     move.w  mm_alchp,a2         Claim channel definition block
           moveq   #$18,d1             24 bytes is mimimum
           jsr     (a2)                Go get the space (exit with address in A0)
           rts                         Return any error codes

*----------------------------------------------------------------------*
* The I/O procedure first checks to see that a valid operation code is *
* present in D0.B, the top 3 bytes are cleared by the IOSS so bytes it *
* is. If the code is invalid, return BAD_PARAMETER, if the code is for *
* an input type operation, return END_OF_FILE, if the code is NOT for  *
* FS_SAVE or IO_SSTRG then return OK otherwise adjust A1.L to point to *
* A1.L + D2.W, return D1.W as D2.W and set no errors. (Follow that ?)  *
*----------------------------------------------------------------------*

dev_io     cmpi.b  #io_sbyte,d0        Check for input operation
           bge.s   not_input           It is not input

error_ef   moveq   #err_ef,d0          Return EOF on all input operations
           rts

error_bp   moveq   #err_bp,d0          BAD PARAMETER for all invalid operations
           rts

not_input  cmpi.b  #io_sbyte,d0        Send a byte, do nothing
           beq.s   exit_ok
           cmpi.b  #io_sstrg,d0        Send a string ?
           beq.s   sstrg_save
           cmpi.b  #fs_save,d0         And save a file require extra code
           bne.s   not_save

sstrg_save adda.w  d2,a1               Adjust the buffer pointer
           move.w  d2,d1               And set the number of bytes sent

exit_ok    moveq   #0,d0               No errors
           rts                         And exit

not_save   cmpi.b  #fs_mdinf,d0        Read media information, EOF
           beq.s   error_ef
           cmpi.b  #fs_headr,d0        Read file header, EOF
           beq.s   error_ef
           cmpi.b  #fs_load,d0         Load a file, EOF
           beq.s   error_ef
           cmpi.b  #fs_check,d0        Try silly codes such as CLS, PAPER etc
           bcs.s   error_bp            Silly ...
           cmpi.b  #fs_posre,d0
           bhi.s   error_bp            Silly ...

rest_ok    bra.s   exit_ok             All the rest, just return ok

           END

